from collections.abc import Iterable
from sys import version_info

from pantherdb import Cursor

from panther.db.connections import db
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import prepare_id_for_query
from panther.exceptions import DatabaseError

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BasePantherDBQuery')


class BasePantherDBQuery(BaseQuery):
    @classmethod
    def _merge(cls, *args, is_mongo: bool = False) -> dict:
        return super()._merge(*args, is_mongo=is_mongo)

    # # # # # Find # # # # #
    @classmethod
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).find_one(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> Cursor:
        cursor = db.session.collection(cls.__name__).find(**cls._merge(_filter, kwargs))
        cursor.response_type = cls._create_model_instance
        cursor.cls = cls
        return cursor

    @classmethod
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).first(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := db.session.collection(cls.__name__).last(**cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def aggregate(cls, *args, **kwargs):
        msg = 'aggregate() does not supported in `PantherDB`.'
        raise DatabaseError(msg) from None

    # # # # # Count # # # # #
    @classmethod
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).count(**cls._merge(_filter, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    async def insert_one(cls, _document: dict | None = None, /, **kwargs) -> Self:
        _document = cls._merge(_document, kwargs)
        cls._validate_data(data=_document)

        document = db.session.collection(cls.__name__).insert_one(**_document)
        return await cls._create_model_instance(document=document)

    @classmethod
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        result = []
        for document in documents:
            prepare_id_for_query(document, is_mongo=False)
            cls._validate_data(data=document)
            inserted_document = db.session.collection(cls.__name__).insert_one(**document)
            document['_id'] = inserted_document['_id']
            result.append(await cls._create_model_instance(document=document))
        return result

    # # # # # Delete # # # # #
    async def delete(self) -> None:
        db.session.collection(self.__class__.__name__).delete_one(_id=self._id)

    @classmethod
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        return db.session.collection(cls.__name__).delete_one(**cls._merge(_filter, kwargs))

    @classmethod
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return db.session.collection(cls.__name__).delete_many(**cls._merge(_filter, kwargs))

    # # # # # Update # # # # #
    async def update(self, _update: dict | None = None, /, **kwargs) -> None:
        document = self._merge(_update, kwargs)
        document.pop('_id', None)
        self._validate_data(data=kwargs, is_updating=True)

        for field, value in document.items():
            if isinstance(value, dict):
                value = type(getattr(self, field))(**value)
            setattr(self, field, value)
        db.session.collection(self.__class__.__name__).update_one({'_id': self._id}, **document)

    @classmethod
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter)
        return db.session.collection(cls.__name__).update_one(_filter, **cls._merge(_update, kwargs))

    @classmethod
    async def update_many(cls, _filter: dict, _data: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter)
        return db.session.collection(cls.__name__).update_many(_filter, **cls._merge(_data, kwargs))



from unittest import IsolatedAsyncioTestCase

from panther import Panther, status
from panther.app import API, GenericAPI
from panther.configs import config
from panther.openapi.urls import url_routing
from panther.openapi.utils import ParseEndpoint
from panther.response import Response
from panther.test import APIClient


@API()
def my_api1():
    return {'detail': 'ok'}


@API()
def my_api2():
    data = {'detail': 'ok'}
    return data


@API()
def my_api3():
    my_data = {'detail': 'ok'}
    return Response(data=my_data)


@API()
async def my_api4():
    return Response(data={'detail': 'ok'})


@API()
async def my_api5():
    return Response(status_code=201)


@API()
async def my_api6():
    return Response(status_code=status.HTTP_207_MULTI_STATUS)


@API()
def my_api7():
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(status_code=my_status)


@API()
def my_api8():
    my_status = 207
    return Response(status_code=my_status)


@API()
def my_api9():
    return Response({'detail': 'ok'})


@API()
def my_api10():
    my_data = {'detail': 'ok'}
    return Response(my_data)


@API()
def my_api11():
    my_data = {'detail': 'ok'}
    return Response(my_data, status_code=207)


@API()
def my_api12():
    my_data = {'detail': 'ok'}
    return Response(my_data, status_code=status.HTTP_207_MULTI_STATUS)


@API()
def my_api13():
    my_data = {'detail': 'ok'}
    my_status = 207
    return Response(my_data, status_code=my_status)


@API()
def my_api14():
    my_data = {'detail': 'ok'}
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(my_data, status_code=my_status)


@API()
def my_api15():
    my_data = {'detail': 'ok'}
    my_status = status.HTTP_207_MULTI_STATUS
    return Response(my_data, my_status)


@API()
def my_api16():
    my_data = {'detail': 'ok'}
    return Response(my_data, status.HTTP_207_MULTI_STATUS)


@API()
def my_api17():
    return Response({}, 207)


class API18(GenericAPI):
    def get(self, *args, **kwargs):
        return {'detail': 'ok'}

    def post(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return my_data

    async def put(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return Response(data=my_data)

    def patch(self, *args, **kwargs):
        my_data = {'detail': 'ok'}
        return Response(data=my_data, status_code=201)

    def delete(self, *args, **kwargs):
        return Response({}, status.HTTP_204_NO_CONTENT)


class TestPanelAPIs(IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls={'swagger': url_routing})
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    async def test_swagger(self):
        response = await self.client.get('/swagger/')
        expected_response = """<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Swagger UI</title>
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/4.16.0/swagger-ui.css" />
</head>
<body>
    <div id="swagger-ui"></div>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/4.16.0/swagger-ui-bundle.js"></script>
    <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist/swagger-ui-standalone-preset.js"></script>
    <script>
    const openapiContent = {"components": {}, "openapi": "3.0.0", "paths": {"/swagger/": {"get": {"responses": {"200": {"content": {"application/json": {"schema": {"properties": {}}}}}}, "summary": null, "tags": ["panther.openapi"], "title": "OpenAPI"}}}};

    const ui = SwaggerUIBundle({
        spec: openapiContent,
        dom_id: '#swagger-ui',
        presets: [
            SwaggerUIBundle.presets.apis,
            SwaggerUIStandalonePreset
        ],
        layout: "StandaloneLayout",
    });
    </script>
</body>
</html>"""
        assert expected_response == response.data

    async def test_my_api1(self):
        parsed = ParseEndpoint(my_api1, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api2(self):
        parsed = ParseEndpoint(my_api2, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api3(self):
        parsed = ParseEndpoint(my_api3, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api4(self):
        parsed = ParseEndpoint(my_api4, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api5(self):
        parsed = ParseEndpoint(my_api5, 'get')
        assert parsed.status_code == 201
        assert parsed.data == {}

    async def test_my_api6(self):
        parsed = ParseEndpoint(my_api6, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {}

    async def test_my_api7(self):
        parsed = ParseEndpoint(my_api7, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {}

    async def test_my_api8(self):
        parsed = ParseEndpoint(my_api8, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {}

    async def test_my_api9(self):
        parsed = ParseEndpoint(my_api9, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api10(self):
        parsed = ParseEndpoint(my_api10, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api11(self):
        parsed = ParseEndpoint(my_api11, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api12(self):
        parsed = ParseEndpoint(my_api12, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api13(self):
        parsed = ParseEndpoint(my_api13, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api14(self):
        parsed = ParseEndpoint(my_api14, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api15(self):
        parsed = ParseEndpoint(my_api15, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api16(self):
        parsed = ParseEndpoint(my_api16, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api17(self):
        parsed = ParseEndpoint(my_api17, 'get')
        assert parsed.status_code == 207
        assert parsed.data == {}

    async def test_my_api15_get(self):
        parsed = ParseEndpoint(API18, 'get')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api15_post(self):
        parsed = ParseEndpoint(API18, 'post')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api15_put(self):
        parsed = ParseEndpoint(API18, 'put')
        assert parsed.status_code == 200
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api15_patch(self):
        parsed = ParseEndpoint(API18, 'patch')
        assert parsed.status_code == 201
        assert parsed.data == {'detail': 'ok'}

    async def test_my_api15_delete(self):
        parsed = ParseEndpoint(API18, 'delete')
        assert parsed.status_code == 204
        assert parsed.data == {}



from unittest import IsolatedAsyncioTestCase

from panther import Panther
from panther.app import API, GenericAPI
from panther.configs import config
from panther.exceptions import PantherError
from panther.permissions import BasePermission
from panther.request import Request
from panther.test import APIClient


class AlwaysDeniedPermission(BasePermission):
    @classmethod
    async def authorization(cls, request: Request) -> bool:
        return False


class NotInheritedPermission:
    @classmethod
    async def authorization(cls, request: Request) -> bool:
        return False


class NotClassMethodPermission(BasePermission):
    async def authorization(cls, request: Request) -> bool:
        return False


class SyncPermission(BasePermission):
    @classmethod
    def authorization(cls, request: Request) -> bool:
        return False


@API()
async def without_permission_api(request: Request):
    return request.user


@API(permissions=[AlwaysDeniedPermission])
async def denied_permission_api(request: Request):
    return request.user


@API(permissions=[NotInheritedPermission])
async def not_inherited_permission_api(request: Request):
    return request.user


urls = {
    'without': without_permission_api,
    'denied-permission': denied_permission_api,
    'not-inherited-permission': not_inherited_permission_api,
}


class TestJWTAuthentication(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        app = Panther(__name__, configs=__name__, urls=urls)
        self.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls):
        config.refresh()

    async def test_without_permission(self):
        res = await self.client.get('without')
        assert res.status_code == 200
        assert res.data is None

    async def test_denied_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('denied-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_not_inherited_permission(self):
        with self.assertNoLogs(level='ERROR'):
            res = await self.client.get('not-inherited-permission')
        assert res.status_code == 403
        assert res.data['detail'] == 'Permission Denied'

    async def test_not_classmethod_permission(self):
        try:

            @API(permissions=[NotClassMethodPermission])
            async def not_classmethod_permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'NotClassMethodPermission.authorization() should be `@classmethod`'
        else:
            assert False

    async def test_not_classmethod_permission_classbased(self):
        try:

            class NotClassMethodPermissionAPI(GenericAPI):
                permissions = [NotClassMethodPermission]
        except PantherError as e:
            assert e.args[0] == 'NotClassMethodPermission.authorization() should be `@classmethod`'
        else:
            assert False

    async def test_sync_permission(self):
        try:

            @API(permissions=[SyncPermission])
            async def sync_permission_api(request: Request):
                return request.user
        except PantherError as e:
            assert e.args[0] == 'SyncPermission.authorization() should be `async`'
        else:
            assert False

    async def test_sync_permission_classbased(self):
        try:

            class SyncPermissionAPI(GenericAPI):
                permissions = [SyncPermission]
        except PantherError as e:
            assert e.args[0] == 'SyncPermission.authorization() should be `async`'
        else:
            assert False



from panther.request import Request
from panther.websocket import Websocket


class BasePermission:
    @classmethod
    async def authorization(cls, request: Request | Websocket) -> bool:
        return True


class AdminPermission(BasePermission):
    @classmethod
    async def authorization(cls, request: Request) -> bool:
        return request.user and getattr(request.user, 'is_admin', False)



import logging

from panther import status
from panther.app import API, GenericAPI
from panther.configs import config
from panther.db.models import BaseUser
from panther.exceptions import AuthenticationAPIError, RedirectAPIError
from panther.panel.middlewares import RedirectToSlashMiddleware
from panther.panel.utils import clean_model_schema, get_models
from panther.permissions import BasePermission
from panther.request import Request
from panther.response import Cookie, RedirectResponse, Response, TemplateResponse

logger = logging.getLogger('panther')


class AdminPanelPermission(BasePermission):
    """We didn't want to change AUTHENTICATION class of user, so we use permission class for this purpose."""

    @classmethod
    async def authorization(cls, request: Request) -> bool:
        from panther.authentications import CookieJWTAuthentication

        try:
            await CookieJWTAuthentication.authentication(request=request)
            return True
        except AuthenticationAPIError:
            raise RedirectAPIError(url=f'login?redirect_to={request.path}')


class LoginView(GenericAPI):
    middlewares = [RedirectToSlashMiddleware]

    def get(self, request: Request):
        return TemplateResponse(name='login.html')

    async def post(self, request: Request):
        from panther.authentications import JWTAuthentication

        user: BaseUser = await config.USER_MODEL.find_one({config.USER_MODEL.USERNAME_FIELD: request.data['username']})
        if user is None:
            logger.debug('User not found.')
            return TemplateResponse(
                name='login.html',
                status_code=status.HTTP_400_BAD_REQUEST,
                context={'error': 'Authentication Error'},
            )
        if user.check_password(password=request.data['password']) is False:
            logger.debug('Password is incorrect.')
            return TemplateResponse(
                name='login.html',
                status_code=status.HTTP_400_BAD_REQUEST,
                context={'error': 'Authentication Error'},
            )
        tokens = await JWTAuthentication.login(user=user)
        return RedirectResponse(
            url=request.query_params.get('redirect_to', '..'),
            status_code=status.HTTP_302_FOUND,
            set_cookies=[
                Cookie(key='access_token', value=tokens['access_token'], max_age=config.JWT_CONFIG.life_time),
                Cookie(key='refresh_token', value=tokens['refresh_token'], max_age=config.JWT_CONFIG.refresh_life_time),
            ],
        )


class HomeView(GenericAPI):
    permissions = [AdminPanelPermission]

    def get(self):
        return TemplateResponse(name='home.html', context={'tables': get_models()})


class TableView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, request: Request, index: int):
        model = config.MODELS[index]
        if data := await model.find():
            data = data
        else:
            data = []

        return TemplateResponse(
            name='table.html',
            context={
                'fields': clean_model_schema(model.schema()),
                'tables': get_models(),
                'records': Response.prepare_data(data),
            },
        )


class CreateView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, request: Request, index: int):
        model = config.MODELS[index]
        return TemplateResponse(
            name='create.html',
            context={
                'fields': clean_model_schema(model.schema()),
                'tables': get_models(),
            },
        )

    async def post(self, request: Request, index: int):
        model = config.MODELS[index]
        validated_data = API.validate_input(model=model, request=request)
        instance = await model.insert_one(validated_data.model_dump())
        if issubclass(model, BaseUser):
            await instance.set_password(password=instance.password)
        return instance


class DetailView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, index: int, document_id: str):
        model = config.MODELS[index]
        obj = await model.find_one_or_raise(id=document_id)
        return TemplateResponse(
            name='detail.html',
            context={'fields': clean_model_schema(model.schema()), 'data': obj.model_dump()},
        )

    async def put(self, request: Request, index: int, document_id: str):
        model = config.MODELS[index]
        validated_data = API.validate_input(model=model, request=request)
        return await model.update_one({'id': document_id}, validated_data.model_dump())

    async def delete(self, index: int, document_id: str):
        model = config.MODELS[index]
        await model.delete_one(id=document_id)
        return Response(status_code=status.HTTP_204_NO_CONTENT)



import ast
import inspect

import pydantic

from panther import status
from panther.serializer import ModelSerializer


class EmptyResponseModel(pydantic.BaseModel):
    pass


class OutputSchema:
    """
    Its values only used in OpenAPI response schema
    """

    def __init__(
        self,
        model: type[ModelSerializer] | type[pydantic.BaseModel] = EmptyResponseModel,
        status_code: int = status.HTTP_200_OK,
    ):
        self.model = model
        self.status_code = status_code


class ParseEndpoint:
    """
    ParseEndpoint parses the endpoint (function-base/ class-base) and finds where it returns
        and extract the `data` and `status_code` values.
    """

    def __init__(self, endpoint, method):
        self.tree = ast.parse(inspect.getsource(endpoint))
        self.method = method

        self.status_code = status.HTTP_200_OK
        self.title = None
        self.data = {}

        self.parse()

    def parse(self):
        for branch in self.tree.body:
            match branch:
                case ast.ClassDef(name=name, body=body):
                    # class ...(GenericAPI):
                    #     def get(...
                    self.title = name
                    for part in body:
                        match part:
                            case ast.FunctionDef(name=name, body=function_body):
                                # def get(...
                                if name == self.method:
                                    self.parse_function(body=function_body)
                                    break
                            case ast.AsyncFunctionDef(name=name, body=function_body):
                                # async def get(...
                                if name == self.method:
                                    self.parse_function(body=function_body)
                                    break

                case ast.FunctionDef(name=name, body=body):
                    # def api(...
                    self.title = name
                    self.parse_function(body=body)

                case ast.AsyncFunctionDef(name=name, body=body):
                    # async def api(...
                    self.title = name
                    self.parse_function(body=body)

    def parse_function(self, body):
        for part in body:
            match part:
                case ast.Return(value=ast.Dict(keys=keys, values=values)):
                    # return {...}
                    self.status_code = 200
                    self.parse_dict_response(keys=keys, values=values)

                case ast.Return(value=ast.Name(id=name)):
                    # return my_data
                    self.status_code = 200
                    for value in self.searching_variable(body=body, name=name):
                        match value:
                            # my_data = {...}
                            case ast.Dict(keys=keys, values=values):
                                self.parse_dict_response(keys=keys, values=values)

                case ast.Return(value=ast.Call(args=args, keywords=keywords, func=func)):
                    if func.id == 'TemplateResponse':
                        return
                    # return Response(...
                    self.parse_response(body=body, args=args, keywords=keywords)

    def parse_dict_response(self, keys, values):
        for k, v in zip(keys, values):
            final_value = None
            match v:
                case ast.Constant(value=value):
                    final_value = value
            self.data[k.value] = final_value

    def parse_response(self, body, args, keywords):
        for keyword in keywords:
            if keyword.arg == 'data':
                self.parse_data(body=body, value=keyword.value)
            if keyword.arg == 'status_code':
                self.parse_status_code(body=body, value=keyword.value)

        for i, arg in enumerate(args):
            if i == 0:  # index 0 is `data`
                self.parse_data(body=body, value=arg)
            elif i == 1:  # index 1 is `status_code`
                self.parse_status_code(body=body, value=arg)

    def parse_status_code(self, body, value):
        match value:
            # return Response(?, status_code=my_status)
            # return Response(?, my_status)
            case ast.Name():
                for inner_value in self.searching_variable(body=body, name=value.id):
                    match inner_value:
                        # my_status = status.HTTP_202_ACCEPTED
                        case ast.Attribute(value=inner_inner_value, attr=attr):
                            if inner_inner_value.id == 'status':
                                self.status_code = getattr(status, attr)
                        # my_status = 202
                        case ast.Constant(value=inner_inner_value):
                            self.status_code = inner_inner_value

            # return Response(?, status_code=status.HTTP_202_ACCEPTED)
            # return Response(?, status.HTTP_202_ACCEPTED)
            case ast.Attribute(value=value, attr=attr):
                if value.id == 'status':
                    self.status_code = getattr(status, attr)
            # return Response(?, status_code=202)
            # return Response(?, 202)
            case ast.Constant(value=value):
                self.status_code = value

    def parse_data(self, body, value):
        match value:
            # return Response(data=my_data, ?)
            # return Response(my_data, ?)
            case ast.Name():
                for value in self.searching_variable(body=body, name=value.id):
                    match value:
                        # my_data = {...}
                        case ast.Dict(keys=keys, values=values):
                            self.parse_dict_response(keys=keys, values=values)

            # return Response(data={...}, ?)
            # return Response({...}, ?)
            case ast.Dict(keys=keys, values=values):
                self.parse_dict_response(keys=keys, values=values)

    def searching_variable(self, body, name):
        for part in body:
            match part:
                case ast.Assign(targets=targets, value=value):
                    for target in targets:
                        match target:
                            case ast.Name(id=inner_name):
                                if inner_name == name:
                                    yield value



import functools
import logging
import traceback
import typing
from collections.abc import Callable
from datetime import timedelta
from typing import Literal

from orjson import JSONDecodeError
from pydantic import BaseModel, ValidationError

from panther._utils import is_function_async
from panther.base_request import BaseRequest
from panther.caching import (
    get_response_from_cache,
    set_response_in_cache,
)
from panther.configs import config
from panther.exceptions import (
    APIError,
    AuthorizationAPIError,
    BadRequestAPIError,
    JSONDecodeAPIError,
    MethodNotAllowedAPIError,
    PantherError,
)
from panther.middlewares import HTTPMiddleware
from panther.openapi import OutputSchema
from panther.permissions import BasePermission
from panther.request import Request
from panther.response import Response
from panther.serializer import ModelSerializer
from panther.throttling import Throttle

__all__ = ('API', 'GenericAPI')

logger = logging.getLogger('panther')


class API:
    """
    methods: Specify the allowed methods.
    input_model: The `request.data` will be validated with this attribute, It will raise an
        `panther.exceptions.BadRequestAPIError` or put the validated data in the `request.validated_data`.
    output_schema: This attribute only used in creation of OpenAPI scheme which is available in `panther.openapi.urls`
        You may want to add its `url` to your urls.
    auth: It will authenticate the user with header of its request or raise an
        `panther.exceptions.AuthenticationAPIError`.
    permissions: List of permissions that will be called sequentially after authentication to authorize the user.
    throttling: It will limit the users' request on a specific (time-window, path)
    cache: Specify the duration of the cache (Will be used only in GET requests).
    middlewares: These middlewares have inner priority than global middlewares.
    """

    func: Callable

    def __init__(
        self,
        *,
        methods: list[Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']] | None = None,
        input_model: type[ModelSerializer] | type[BaseModel] | None = None,
        output_schema: OutputSchema | None = None,
        auth: bool = False,
        permissions: list[type[BasePermission]] | None = None,
        throttling: Throttle | None = None,
        cache: timedelta | None = None,
        middlewares: list[type[HTTPMiddleware]] | None = None,
        **kwargs,
    ):
        self.methods = {m.upper() for m in methods} if methods else {'GET', 'POST', 'PUT', 'PATCH', 'DELETE'}
        self.input_model = input_model
        self.output_schema = output_schema
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.middlewares: list[[HTTPMiddleware]] | None = middlewares
        self.request: Request | None = None
        if kwargs.pop('output_model', None):
            deprecation_message = (
                traceback.format_stack(limit=2)[0]
                + '\nThe `output_model` argument has been removed in Panther v5 and is no longer available.'
                '\nPlease update your code to use the new approach. More info: '
                'https://pantherpy.github.io/open_api/'
            )
            raise PantherError(deprecation_message)
        if kwargs.pop('cache_exp_time', None):
            deprecation_message = (
                traceback.format_stack(limit=2)[0]
                + '\nThe `cache_exp_time` argument has been removed in Panther v5 and is no longer available.'
                '\nYou may want to use `cache` instead.'
            )
            raise PantherError(deprecation_message)
        # Validate Cache
        if self.cache and not isinstance(self.cache, timedelta):
            deprecation_message = (
                traceback.format_stack(limit=2)[0] + '\nThe `cache` argument has been changed in Panther v5, '
                'it should be an instance of `datetime.timedelta()`.'
            )
            raise PantherError(deprecation_message)
        assert self.cache is None or isinstance(self.cache, timedelta)
        # Validate Permissions
        for perm in self.permissions:
            if is_function_async(perm.authorization) is False:
                msg = f'{perm.__name__}.authorization() should be `async`'
                logger.error(msg)
                raise PantherError(msg)
            if type(perm.authorization).__name__ != 'method':
                msg = f'{perm.__name__}.authorization() should be `@classmethod`'
                logger.error(msg)
                raise PantherError(msg)
        # Check kwargs
        if kwargs:
            msg = f'Unknown kwargs: {kwargs.keys()}'
            logger.error(msg)
            raise PantherError(msg)

    def __call__(self, func):
        self.func = func
        self.is_function_async = is_function_async(self.func)
        self.function_annotations = {
            k: v for k, v in func.__annotations__.items() if v in {BaseRequest, Request, bool, int}
        }

        @functools.wraps(func)
        async def wrapper(request: Request) -> Response:
            chained_func = self.handle_endpoint
            if self.middlewares:
                for middleware in reversed(self.middlewares):
                    chained_func = middleware(chained_func)
            return await chained_func(request=request)

        # Store attributes on the function, so have the same behaviour as class-based (useful in `openapi.view.OpenAPI`)
        wrapper.auth = self.auth
        wrapper.methods = self.methods
        wrapper.permissions = self.permissions
        wrapper.input_model = self.input_model
        wrapper.output_schema = self.output_schema
        return wrapper

    async def handle_endpoint(self, request: Request) -> Response:
        self.request = request

        # 1. Check Method
        if self.request.method not in self.methods:
            raise MethodNotAllowedAPIError

        # 2. Authentication
        if self.auth:
            if not config.AUTHENTICATION:
                logger.critical('"AUTHENTICATION" has not been set in configs')
                raise APIError
            self.request.user = await config.AUTHENTICATION.authentication(self.request)

        # 3. Permissions
        for perm in self.permissions:
            if await perm.authorization(self.request) is False:
                raise AuthorizationAPIError

        # 4. Throttle
        if throttling := self.throttling or config.THROTTLING:
            await throttling.check_and_increment(request=self.request)

        # 5. Validate Input
        if self.input_model and self.request.method in {'POST', 'PUT', 'PATCH'}:
            self.request.validated_data = self.validate_input(model=self.input_model, request=self.request)

        # 6. Get Cached Response
        if self.cache and self.request.method == 'GET':
            if cached := await get_response_from_cache(request=self.request, duration=self.cache):
                return Response(data=cached.data, headers=cached.headers, status_code=cached.status_code)

        # 7. Put PathVariables and Request(If User Wants It) In kwargs
        kwargs = self.request.clean_parameters(self.function_annotations)

        # 8. Call Endpoint
        if self.is_function_async:
            response = await self.func(**kwargs)
        else:
            response = self.func(**kwargs)

        # 9. Clean Response
        if not isinstance(response, Response):
            response = Response(data=response)
        if response.pagination:
            response.data = await response.pagination.template(response.data)

        # 10. Set New Response To Cache
        if self.cache and self.request.method == 'GET':
            await set_response_in_cache(request=self.request, response=response, duration=self.cache)

        return response

    @classmethod
    def validate_input(cls, model, request: Request):
        if isinstance(request.data, bytes):
            raise BadRequestAPIError(detail='Content-Type is not valid')
        if request.data is None:
            raise BadRequestAPIError(detail='Request body is required')
        try:
            # `request` will be ignored in regular `BaseModel`
            return model(**request.data, request=request)
        except ValidationError as validation_error:
            error = {'.'.join(str(loc) for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise BadRequestAPIError(detail=error)
        except JSONDecodeError:
            raise JSONDecodeAPIError


class MetaGenericAPI(type):
    def __new__(cls, cls_name: str, bases: tuple[type[typing.Any], ...], namespace: dict[str, typing.Any], **kwargs):
        if cls_name == 'GenericAPI':
            return super().__new__(cls, cls_name, bases, namespace)
        if 'output_model' in namespace:
            deprecation_message = (
                traceback.format_stack(limit=2)[0]
                + '\nThe `output_model` argument has been removed in Panther v5 and is no longer available.'
                '\nPlease update your code to use the new approach. More info: '
                'https://pantherpy.github.io/open_api/'
            )
            raise PantherError(deprecation_message)
        return super().__new__(cls, cls_name, bases, namespace)


class GenericAPI(metaclass=MetaGenericAPI):
    """
    Check out the documentation of `panther.app.API()`.
    """

    input_model: type[ModelSerializer] | type[BaseModel] | None = None
    output_schema: OutputSchema | None = None
    auth: bool = False
    permissions: list[type[BasePermission]] | None = None
    throttling: Throttle | None = None
    cache: timedelta | None = None
    middlewares: list[HTTPMiddleware] | None = None

    def __init_subclass__(cls, **kwargs):
        # Creating API instance to validate the attributes.
        API(
            input_model=cls.input_model,
            output_schema=cls.output_schema,
            auth=cls.auth,
            permissions=cls.permissions,
            throttling=cls.throttling,
            cache=cls.cache,
            middlewares=cls.middlewares,
        )

    async def get(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def post(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def put(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def patch(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def delete(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def call_method(self, request: Request):
        match request.method:
            case 'GET':
                func = self.get
            case 'POST':
                func = self.post
            case 'PUT':
                func = self.put
            case 'PATCH':
                func = self.patch
            case 'DELETE':
                func = self.delete
            case _:
                raise MethodNotAllowedAPIError

        return await API(
            input_model=self.input_model,
            output_schema=self.output_schema,
            auth=self.auth,
            permissions=self.permissions,
            throttling=self.throttling,
            cache=self.cache,
            middlewares=self.middlewares,
        )(func)(request=request)



import contextlib
from typing import List
from unittest import IsolatedAsyncioTestCase

import faker
import pytest
from pydantic import BaseModel

from panther import Panther
from panther.configs import config
from panther.db import Model
from panther.db.connections import db
from panther.exceptions import DatabaseError

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson

f = faker.Faker()


class Book(Model):
    name: str


class BookDetail(BaseModel):
    detail: str
    book: Book
    more_books: list[list[Book]]


class Author(Model):
    name: str
    books: list[Book]
    books2: List[Book]
    book: Book
    book2: None | Book = None
    book_detail: dict
    our_book_detail: BookDetail


class InvalidModel(Model):
    new_book: dict[str, Book]


class _BaseDatabaseTestCase:
    async def test_insert_one(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=book,
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book == book

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_many(self):
        book = await Book.insert_one(name='my_test')
        authors = await Author.insert_many(
            [
                {
                    'name': 'ali',
                    'books': [book],
                    'books2': [book, book],
                    'book': book,
                    'book2': None,
                    'book_detail': {'book1': book},
                    'our_book_detail': BookDetail(detail='ok', book=book, more_books=[[book, book]]),
                },
                {
                    'name': 'ali',
                    'books': [book],
                    'books2': [book, book],
                    'book': book,
                    'book2': None,
                    'book_detail': {'book1': book},
                    'our_book_detail': BookDetail(detail='ok', book=book, more_books=[[book, book]]),
                },
            ],
        )

        for author in authors:
            assert isinstance(book, Book)
            assert book.id
            assert book.name == 'my_test'

            assert author.name == 'ali'
            assert isinstance(author.books, list)
            assert len(author.books) == 1
            assert author.books[0] == book
            assert author.books[0]

            assert isinstance(author.books2, list)
            assert len(author.books2) == 2
            assert author.books2[0] == book
            assert author.books2[1] == book

            assert isinstance(author.book, Book)
            assert author.book == book

            assert author.book2 is None

            assert isinstance(author.book_detail, dict)
            assert list(author.book_detail.keys()) == ['book1']
            assert author.book_detail['book1'] == book.id  # Known Issue

            assert isinstance(author.our_book_detail, BookDetail)
            assert author.our_book_detail.detail == 'ok'
            assert author.our_book_detail.book == book
            assert isinstance(author.our_book_detail.more_books, list)
            assert len(author.our_book_detail.more_books) == 1
            assert isinstance(author.our_book_detail.more_books[0], list)
            assert len(author.our_book_detail.more_books[0]) == 2
            assert author.our_book_detail.more_books[0][0] == book
            assert author.our_book_detail.more_books[0][1] == book

    async def test_find_one(self):
        book = await Book.insert_one(name='my_test')
        await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=book,
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )
        author = await Author.find_one(name='ali')

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book == book

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_find(self):
        book = await Book.insert_one(name='my_test')
        await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=book,
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )
        _authors = await Author.find(name='ali')
        authors = [i for i in _authors]
        assert len(authors) == 1
        author = authors[0]

        assert isinstance(book, Book)
        assert book.id
        assert book.name == 'my_test'

        assert author.name == 'ali'
        assert isinstance(author.books, list)
        assert len(author.books) == 1
        assert author.books[0] == book
        assert author.books[0]

        assert isinstance(author.books2, list)
        assert len(author.books2) == 2
        assert author.books2[0] == book
        assert author.books2[1] == book

        assert isinstance(author.book, Book)
        assert author.book == book

        assert author.book2 is None

        assert isinstance(author.book_detail, dict)
        assert list(author.book_detail.keys()) == ['book1']
        assert author.book_detail['book1'] == book.id  # Known Issue

        assert isinstance(author.our_book_detail, BookDetail)
        assert author.our_book_detail.detail == 'ok'
        assert author.our_book_detail.book == book
        assert isinstance(author.our_book_detail.more_books, list)
        assert len(author.our_book_detail.more_books) == 1
        assert isinstance(author.our_book_detail.more_books[0], list)
        assert len(author.our_book_detail.more_books[0]) == 2
        assert author.our_book_detail.more_books[0][0] == book
        assert author.our_book_detail.more_books[0][1] == book

    async def test_insert_one_invalid_model(self):
        book = await Book.insert_one(name='my_test')
        try:
            await InvalidModel.insert_one(new_book={'book': book})
        except DatabaseError as e:
            assert (
                e.args[0]
                == 'Panther does not support dict[str, tests.test_database_advance.Book] as a field type for unwrapping.'
            )


@pytest.mark.mongodb
class TestMongoDB(_BaseDatabaseTestCase, IsolatedAsyncioTestCase):
    DB_NAME = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global DATABASE
        DATABASE = {
            'engine': {
                'class': 'panther.db.connections.MongoDBConnection',
                'host': f'mongodb://127.0.0.1:27017/{cls.DB_NAME}',
            },
        }

    def setUp(self):
        Panther(__name__, configs=__name__, urls={})

    def tearDown(self) -> None:
        db.session.drop_collection('Book')
        db.session.drop_collection('Author')

    @classmethod
    def tearDownClass(cls) -> None:
        config.refresh()

    async def test_insert_one_raw_document(self):
        book = await Book.insert_one(name='my_test')
        author = await Author.insert_one(
            name='ali',
            books=[book],
            books2=[book, book],
            book=book,
            book2=None,
            book_detail={'book1': book},
            our_book_detail=BookDetail(detail='ok', book=book, more_books=[[book, book]]),
        )

        document = await db.session['Author'].find_one()
        assert isinstance(document, dict)
        assert list(document.keys()) == [
            '_id',
            'name',
            'books',
            'books2',
            'book',
            'book2',
            'book_detail',
            'our_book_detail',
        ]

        assert document['_id'] == author._id

        assert document['name'] == 'ali'

        assert isinstance(document['books'], list)
        assert len(document['books']) == 1
        assert isinstance(document['books'][0], bson.ObjectId)
        assert document['books'][0] == book._id

        assert isinstance(document['books2'], list)
        assert len(document['books2']) == 2
        assert isinstance(document['books2'][0], bson.ObjectId)
        assert document['books2'][0] == book._id
        assert isinstance(document['books2'][1], bson.ObjectId)
        assert document['books2'][1] == book._id

        assert isinstance(document['book'], bson.ObjectId)
        assert document['book'] == book._id

        assert document['book2'] is None

        assert isinstance(document['book_detail'], dict)
        assert list(document['book_detail'].keys()) == ['book1']
        assert isinstance(document['book_detail']['book1'], bson.ObjectId)
        assert document['book_detail']['book1'] == book._id

        assert isinstance(document['our_book_detail'], dict)
        assert list(document['our_book_detail'].keys()) == ['detail', 'book', 'more_books']
        assert document['our_book_detail']['detail'] == 'ok'
        assert isinstance(document['our_book_detail']['book'], bson.ObjectId)
        assert document['our_book_detail']['book'] == book._id
        assert isinstance(document['our_book_detail']['more_books'], list)
        assert len(document['our_book_detail']['more_books']) == 1
        assert isinstance(document['our_book_detail']['more_books'][0], list)
        assert len(document['our_book_detail']['more_books'][0]) == 2
        assert isinstance(document['our_book_detail']['more_books'][0][0], bson.ObjectId)
        assert document['our_book_detail']['more_books'][0][0] == book._id
        assert isinstance(document['our_book_detail']['more_books'][0][1], bson.ObjectId)
        assert document['our_book_detail']['more_books'][0][1] == book._id



from __future__ import annotations

import types
import typing
from sys import version_info
from typing import Any, Union, get_args, get_origin

from pydantic import BaseModel, ValidationError

from panther.db.connections import db
from panther.db.cursor import Cursor
from panther.db.models import Model
from panther.db.queries.base_queries import BaseQuery
from panther.db.utils import prepare_id_for_query
from panther.exceptions import DatabaseError

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

try:
    from bson.codec_options import CodecOptions
    from pymongo.results import InsertManyResult, InsertOneResult
except ImportError:
    # MongoDB-related libraries are not required by default.
    # If the user intends to use MongoDB, they must install the required dependencies explicitly.
    # This will be enforced in `panther.db.connections.MongoDBConnection.init`.
    CodecOptions = type('CodecOptions', (), {})
    InsertOneResult = type('InsertOneResult', (), {})
    InsertManyResult = type('InsertManyResult', (), {})

if version_info >= (3, 11):
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BaseMongoDBQuery')


def get_annotation_type(annotation: Any) -> type | None:
    """
    Extracts the underlying, non-optional type from a type annotation.
    Handles basic types, Pydantic BaseModels, lists, and unions (optionals).
    Returns None if no single underlying type can be determined (e.g., for list[NoneType]).
    Raises DatabaseError for unsupported annotations.
    """
    origin = get_origin(annotation)

    # Handle list[T] and Union[T, None] (T | None or typing.Union[T, None])
    if origin is list or origin is types.UnionType or origin is Union:
        # Extracts the first non-None type from a tuple of type arguments.
        for arg in get_args(annotation):
            if arg is not type(None):
                return arg
        return None

    # Handle basic types (str, int, bool, dict) and Pydantic BaseModel subclasses
    if isinstance(annotation, type) and (annotation in (str, int, bool, dict) or issubclass(annotation, BaseModel)):
        return annotation

    raise DatabaseError(f'Panther does not support {annotation} as a field type for unwrapping.')


class BaseMongoDBQuery(BaseQuery):
    @classmethod
    def _merge(cls, *args, is_mongo: bool = True) -> dict:
        return super()._merge(*args, is_mongo=is_mongo)

    # TODO: https://jira.mongodb.org/browse/PYTHON-4192
    # @classmethod
    # def collection(cls):
    #     return db.session.get_collection(name=cls.__name__, codec_options=CodecOptions(document_class=cls))

    @classmethod
    async def _create_list(cls, field_type: type, value: Any) -> Any:
        # `field_type` is the expected type of items in the list (e.g., int, Model, list[str])
        # `value` is a single item from the input list that needs processing.

        # Handles list[list[int]], list[dict[str,int]] etc.
        if isinstance(field_type, (types.GenericAlias, typing._GenericAlias)):
            element_type = get_annotation_type(field_type)  # Unwrap further (e.g. list[str] -> str)
            if element_type is None:
                raise DatabaseError(f'Cannot determine element type for generic list item: {field_type}')
            if not isinstance(value, list):  # Or check if iterable, matching the structure
                raise DatabaseError(f'Expected a list for nested generic type {field_type}, got {type(value)}')
            return [await cls._create_list(field_type=element_type, value=item) for item in value]

        # Make sure Model condition is before BaseModel.
        if isinstance(field_type, type) and issubclass(field_type, Model):
            # `value` is assumed to be an ID for the Model instance.
            return await field_type.first(id=value)

        if isinstance(field_type, type) and issubclass(field_type, BaseModel):
            if not isinstance(value, dict):
                raise DatabaseError(f'Expected a dictionary for BaseModel {field_type.__name__}, got {type(value)}')

            return {
                field_name: await cls._create_field(model=field_type, field_name=field_name, value=value[field_name])
                for field_name in value
            }

        # Base case: value is a primitive type (str, int, etc.)
        return value

    @classmethod
    async def _create_field(cls, model: type, field_name: str, value: Any) -> Any:
        # Handle primary key field directly
        if field_name == '_id':
            return value

        if field_name not in model.model_fields:
            # Field from input data is not defined in the model.
            # Pydantic's `extra` config on the model will handle this upon instantiation.
            return value

        field_annotation = model.model_fields[field_name].annotation
        unwrapped_type = get_annotation_type(field_annotation)

        if unwrapped_type is None:
            raise DatabaseError(
                f"Could not determine a valid underlying type for field '{field_name}' "
                f'with annotation {field_annotation} in model {model.__name__}.',
            )

        if get_origin(field_annotation) is list:
            # Or check for general iterables if applicable
            if not isinstance(value, list):
                raise DatabaseError(
                    f"Field '{field_name}' expects a list, got {type(value)} for model {model.__name__}",
                )
            return [await cls._create_list(field_type=unwrapped_type, value=item) for item in value]

        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, Model):
            if obj := await unwrapped_type.first(id=value):
                return obj.model_dump(by_alias=True)
            return None

        if isinstance(unwrapped_type, type) and issubclass(unwrapped_type, BaseModel):
            if not isinstance(value, dict):
                raise DatabaseError(
                    f"Field '{field_name}' expects a dictionary for BaseModel {unwrapped_type.__name__}, "
                    f'got {type(value)} in model {model.__name__}',
                )
            return {
                nested_field_name: await cls._create_field(
                    model=unwrapped_type,
                    field_name=nested_field_name,
                    value=value[nested_field_name],
                )
                for nested_field_name in unwrapped_type.model_fields
                if nested_field_name in value
            }

        return value

    @classmethod
    async def _create_model_instance(cls, document: dict) -> Self:
        """Prepares document and creates an instance of the model."""
        processed_document = {
            field_name: await cls._create_field(model=cls, field_name=field_name, value=field_value)
            for field_name, field_value in document.items()
        }
        try:
            return cls(**processed_document)
        except ValidationError as validation_error:
            error = cls._clean_error_message(validation_error=validation_error)
            raise DatabaseError(error) from validation_error

    @classmethod
    def clean_value(cls, field: str | None, value: Any) -> dict[str, Any] | list[Any]:
        match value:
            case None:
                return None
            case Model() as model:
                if model.id is None:
                    raise DatabaseError(f'Model instance{" in " + field if field else ""} has no ID.')
                # We save full object because user didn't specify the type.
                return model._id
            case BaseModel() as model:
                return {
                    field_name: cls.clean_value(field=field_name, value=getattr(model, field_name))
                    for field_name in model.model_fields
                }
            case dict() as d:
                return {k: cls.clean_value(field=k, value=v) for k, v in d.items()}
            case list() as l:
                return [cls.clean_value(field=None, value=item) for item in l]

        return value

    # # # # # Find # # # # #
    @classmethod
    async def find_one(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        if document := await db.session[cls.__name__].find_one(cls._merge(_filter, kwargs)):
            return await cls._create_model_instance(document=document)
        return None

    @classmethod
    async def find(cls, _filter: dict | None = None, /, **kwargs) -> Cursor:
        return Cursor(cls=cls, collection=db.session[cls.__name__].delegate, filter=cls._merge(_filter, kwargs))

    @classmethod
    async def first(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        async for result in cursor.sort('_id', 1).limit(-1):
            return result
        return None

    @classmethod
    async def last(cls, _filter: dict | None = None, /, **kwargs) -> Self | None:
        cursor = await cls.find(_filter, **kwargs)
        async for result in cursor.sort('_id', -1).limit(-1):
            return result
        return None

    @classmethod
    async def aggregate(cls, pipeline: Sequence[dict]) -> Iterable[dict]:
        return await db.session[cls.__name__].aggregate(pipeline).to_list(None)

    # # # # # Count # # # # #
    @classmethod
    async def count(cls, _filter: dict | None = None, /, **kwargs) -> int:
        return await db.session[cls.__name__].count_documents(cls._merge(_filter, kwargs))

    # # # # # Insert # # # # #
    @classmethod
    async def insert_one(cls, _document: dict | None = None, /, **kwargs) -> Self:
        document = cls._merge(_document, kwargs)
        cls._validate_data(data=document)
        final_document = {field: cls.clean_value(field=field, value=value) for field, value in document.items()}
        result = await cls._create_model_instance(document=final_document)
        insert_one_result: InsertOneResult = await db.session[cls.__name__].insert_one(final_document)
        result.id = insert_one_result.inserted_id
        return result

    @classmethod
    async def insert_many(cls, documents: Iterable[dict]) -> list[Self]:
        final_documents = []
        results = []
        for document in documents:
            prepare_id_for_query(document, is_mongo=True)
            cls._validate_data(data=document)
            cleaned_document = {field: cls.clean_value(field=field, value=value) for field, value in document.items()}
            final_documents.append(cleaned_document)
            results.append(await cls._create_model_instance(document=cleaned_document))
        insert_many_result: InsertManyResult = await db.session[cls.__name__].insert_many(final_documents)
        for obj, _id in zip(results, insert_many_result.inserted_ids):
            obj.id = _id
        return results

    # # # # # Delete # # # # #
    async def delete(self) -> None:
        await db.session[self.__class__.__name__].delete_one({'_id': self._id})

    @classmethod
    async def delete_one(cls, _filter: dict | None = None, /, **kwargs) -> bool:
        result = await db.session[cls.__name__].delete_one(cls._merge(_filter, kwargs))
        return bool(result.deleted_count)

    @classmethod
    async def delete_many(cls, _filter: dict | None = None, /, **kwargs) -> int:
        result = await db.session[cls.__name__].delete_many(cls._merge(_filter, kwargs))
        return result.deleted_count

    # # # # # Update # # # # #
    async def update(self, _update: dict | None = None, /, **kwargs) -> None:
        merged_update_query = self._merge(_update, kwargs)
        merged_update_query.pop('_id', None)

        self._validate_data(data=merged_update_query, is_updating=True)

        update_query = {}
        for field, value in merged_update_query.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                update_query['$set'] = update_query.get('$set', {})
                update_query['$set'][field] = value
                if isinstance(value, dict):
                    value = type(getattr(self, field))(**value)
                setattr(self, field, value)

        await db.session[self.__class__.__name__].update_one({'_id': self._id}, update_query)

    @classmethod
    async def update_one(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> bool:
        prepare_id_for_query(_filter, is_mongo=True)
        merged_update_query = cls._merge(_update, kwargs)

        update_query = {}
        for field, value in merged_update_query.items():
            if field.startswith('$'):
                update_query[field] = value
            else:
                update_query['$set'] = update_query.get('$set', {})
                update_query['$set'][field] = value

        result = await db.session[cls.__name__].update_one(_filter, update_query)
        return bool(result.matched_count)

    @classmethod
    async def update_many(cls, _filter: dict, _update: dict | None = None, /, **kwargs) -> int:
        prepare_id_for_query(_filter, is_mongo=True)
        update_fields = {'$set': cls._merge(_update, kwargs)}

        result = await db.session[cls.__name__].update_many(_filter, update_fields)
        return result.modified_count



import contextlib
import logging

from pantherdb import Cursor as PantherDBCursor

from panther import status
from panther.app import GenericAPI
from panther.configs import config
from panther.db import Model
from panther.db.cursor import Cursor
from panther.exceptions import APIError
from panther.pagination import Pagination
from panther.request import Request
from panther.response import Response
from panther.serializer import ModelSerializer

with contextlib.suppress(ImportError):
    # Only required if user wants to use mongodb
    import bson

logger = logging.getLogger('panther')


class ObjectRequired:
    def _check_object(self, instance):
        if instance and issubclass(type(instance), Model) is False:
            logger.critical(f'`{self.__class__.__name__}.object()` should return instance of a Model --> `find_one()`')
            raise APIError

    async def object(self, request: Request, **kwargs):
        """
        Used in `RetrieveAPI`, `UpdateAPI`, `DeleteAPI`
        """
        logger.error(f'`object()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)


class CursorRequired:
    def _check_cursor(self, cursor):
        if isinstance(cursor, (Cursor, PantherDBCursor)) is False:
            logger.critical(f'`{self.__class__.__name__}.cursor()` should return a Cursor --> `find()`')
            raise APIError

    async def cursor(self, request: Request, **kwargs) -> Cursor | PantherDBCursor:
        """
        Used in `ListAPI`
        Should return `.find()`
        """
        logger.error(f'`cursor()` method is not implemented in {self.__class__} .')
        raise APIError(status_code=status.HTTP_501_NOT_IMPLEMENTED)


class RetrieveAPI(GenericAPI, ObjectRequired):
    async def get(self, request: Request, **kwargs):
        instance = await self.object(request=request, **kwargs)
        self._check_object(instance)

        return Response(data=instance, status_code=status.HTTP_200_OK)


class ListAPI(GenericAPI, CursorRequired):
    sort_fields: list[str]
    search_fields: list[str]
    filter_fields: list[str]
    pagination: type[Pagination]

    async def get(self, request: Request, **kwargs):
        cursor, pagination = await self.prepare_cursor(request=request, **kwargs)
        return Response(data=cursor, pagination=pagination, status_code=status.HTTP_200_OK)

    async def prepare_cursor(self, request: Request, **kwargs) -> tuple[Cursor | PantherDBCursor, Pagination | None]:
        cursor = await self.cursor(request=request, **kwargs)
        self._check_cursor(cursor)

        query = {}
        query |= self.process_filters(query_params=request.query_params, cursor=cursor)
        query |= self.process_search(query_params=request.query_params)

        if query:
            cursor = await cursor.cls.find(cursor.filter | query)

        if sort := self.process_sort(query_params=request.query_params):
            cursor = cursor.sort(sort)

        if pagination := self.process_pagination(query_params=request.query_params, cursor=cursor):
            cursor = pagination.paginate()

        return cursor, pagination

    def process_filters(self, query_params: dict, cursor: Cursor | PantherDBCursor) -> dict:
        _filter = {}
        if hasattr(self, 'filter_fields'):
            for field in self.filter_fields:
                if field in query_params:
                    if config.DATABASE.__class__.__name__ == 'MongoDBConnection':
                        with contextlib.suppress(Exception):
                            # Change type of the value if it is ObjectId
                            if cursor.cls.model_fields[field].metadata[0].func.__name__ == 'validate_object_id':
                                _filter[field] = bson.ObjectId(query_params[field])
                                continue
                    _filter[field] = query_params[field]
        return _filter

    def process_search(self, query_params: dict) -> dict:
        if hasattr(self, 'search_fields') and 'search' in query_params:
            value = query_params['search']
            if config.DATABASE.__class__.__name__ == 'MongoDBConnection':
                if search := [{field: {'$regex': value}} for field in self.search_fields]:
                    return {'$or': search}
            else:
                logger.warning(f'`?search={value} does not work well while using `PantherDB` as Database')
                return {field: value for field in self.search_fields}
        return {}

    def process_sort(self, query_params: dict) -> list:
        if hasattr(self, 'sort_fields') and 'sort' in query_params:
            return [
                (field, -1 if param[0] == '-' else 1)
                for field in self.sort_fields
                for param in query_params['sort'].split(',')
                if field == param.removeprefix('-')
            ]

    def process_pagination(self, query_params: dict, cursor: Cursor | PantherDBCursor) -> Pagination | None:
        if hasattr(self, 'pagination'):
            return self.pagination(query_params=query_params, cursor=cursor)


class CreateAPI(GenericAPI):
    input_model: type[ModelSerializer]

    async def post(self, request: Request, **kwargs):
        instance = await request.validated_data.create(
            validated_data={
                field: getattr(request.validated_data, field)
                for field in request.validated_data.model_fields_set
                if field != 'request'
            },
        )
        return Response(data=instance, status_code=status.HTTP_201_CREATED)


class UpdateAPI(GenericAPI, ObjectRequired):
    input_model: type[ModelSerializer]

    async def put(self, request: Request, **kwargs):
        instance = await self.object(request=request, **kwargs)
        self._check_object(instance)

        await request.validated_data.update(
            instance=instance,
            validated_data=request.validated_data.model_dump(by_alias=True),
        )
        return Response(data=instance, status_code=status.HTTP_200_OK)

    async def patch(self, request: Request, **kwargs):
        instance = await self.object(request=request, **kwargs)
        self._check_object(instance)

        await request.validated_data.partial_update(
            instance=instance,
            validated_data=request.validated_data.model_dump(exclude_none=True, by_alias=True),
        )
        return Response(data=instance, status_code=status.HTTP_200_OK)


class DeleteAPI(GenericAPI, ObjectRequired):
    async def pre_delete(self, instance, request: Request, **kwargs):
        pass

    async def post_delete(self, instance, request: Request, **kwargs):
        pass

    async def delete(self, request: Request, **kwargs):
        instance = await self.object(request=request, **kwargs)
        self._check_object(instance)

        await self.pre_delete(instance, request=request, **kwargs)
        await instance.delete()
        await self.post_delete(instance, request=request, **kwargs)

        return Response(status_code=status.HTTP_204_NO_CONTENT)


class ListCreateAPI(CreateAPI, ListAPI):
    pass


class UpdateDeleteAPI(UpdateAPI, DeleteAPI):
    pass


class RetrieveUpdateDeleteAPI(RetrieveAPI, UpdateAPI, DeleteAPI):
    pass



import types

from panther.app import GenericAPI
from panther.configs import config
from panther.openapi.utils import ParseEndpoint
from panther.response import TemplateResponse


class OpenAPI(GenericAPI):
    @classmethod
    def get_content(cls, endpoint, method):
        parsed = ParseEndpoint(endpoint=endpoint, method=method)

        if endpoint.output_schema:
            status_code = endpoint.output_schema.status_code
            schema = endpoint.output_schema.model.schema()
        else:
            status_code = parsed.status_code
            schema = {'properties': {k: {'default': v} for k, v in parsed.data.items()}}

        responses = {}
        if schema:
            responses = {'responses': {status_code: {'content': {'application/json': {'schema': schema}}}}}
        request_body = {}
        if endpoint.input_model and method in ['post', 'put', 'patch']:
            request_body = {
                'requestBody': {
                    'required': True,
                    'content': {
                        'application/json': {'schema': endpoint.input_model.schema() if endpoint.input_model else {}},
                    },
                },
            }

        content = (
            {
                'title': parsed.title,
                'summary': endpoint.__doc__,
                'tags': ['.'.join(endpoint.__module__.rsplit('.')[:-1]) or endpoint.__module__],
            }
            | responses
            | request_body
        )
        return {method: content}

    def get(self):
        paths = {}
        # TODO:
        #   Try to process the endpoint with `ast` if output_schema is None
        #   Create Component for output_schema.model and input_model
        #
        for url, endpoint in config.FLAT_URLS.items():
            if url == '':
                url = '/'
            if not url.startswith('/'):
                url = f'/{url}'
            paths[url] = {}

            if isinstance(endpoint, types.FunctionType):
                methods = endpoint.methods
                if 'POST' in methods:
                    paths[url] |= self.get_content(endpoint, 'post')
                if 'GET' in methods:
                    paths[url] |= self.get_content(endpoint, 'get')
                if 'PUT' in methods:
                    paths[url] |= self.get_content(endpoint, 'put')
                if 'PATCH' in methods:
                    paths[url] |= self.get_content(endpoint, 'patch')
                if 'DELETE' in methods:
                    paths[url] |= self.get_content(endpoint, 'delete')
            else:
                if endpoint.post is not GenericAPI.post:
                    paths[url] |= self.get_content(endpoint, 'post')
                if endpoint.get is not GenericAPI.get:
                    paths[url] |= self.get_content(endpoint, 'get')
                if endpoint.put is not GenericAPI.put:
                    paths[url] |= self.get_content(endpoint, 'put')
                if endpoint.patch is not GenericAPI.patch:
                    paths[url] |= self.get_content(endpoint, 'patch')
                if endpoint.delete is not GenericAPI.delete:
                    paths[url] |= self.get_content(endpoint, 'delete')

        openapi_content = {'openapi': '3.0.0', 'paths': paths, 'components': {}}
        return TemplateResponse(name='openapi.html', context={'openapi_content': openapi_content})



from functools import cached_property

from pydantic import BaseModel, field_validator, model_serializer

from panther import status
from panther.exceptions import APIError


class File(BaseModel):
    file_name: str
    content_type: str
    file: bytes

    @cached_property
    def size(self):
        return len(self.file)

    def save(self) -> str:
        if hasattr(self, '_file_name'):
            return self._file_name

        self._file_name = self.file_name
        # TODO: check for duplication
        with open(self._file_name, 'wb') as file:
            file.write(self.file)

        return self.file_name

    @model_serializer(mode='wrap')
    def _serialize(self, handler):
        result = handler(self)
        result['path'] = self.save()
        return result

    def __repr__(self) -> str:
        return f'{self.__repr_name__()}(file_name={self.file_name}, content_type={self.content_type})'

    __str__ = __repr__


class Image(File):
    @field_validator('content_type')
    @classmethod
    def validate_content_type(cls, content_type: str) -> str:
        if not content_type.startswith('image/'):
            msg = f"{content_type} is not a valid image 'content_type'"
            raise APIError(detail=msg, status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE)
        return content_type



from datetime import datetime

from panther import version
from panther.utils import generate_secret_key

apis_py = """from datetime import datetime

from app.throttling import InfoThrottle

from panther import status, version
from panther.app import API
from panther.request import Request
from panther.response import Response
from panther.utils import timezone_now


@API()
async def hello_world_api():
    return {'detail': 'Hello World'}


@API(cache=True, throttling=InfoThrottle)
async def info_api(request: Request):
    data = {
        'panther_version': version(),
        'method': request.method,
        'query_params': request.query_params,
        'datetime_now': timezone_now().isoformat(),
        'user_agent': request.headers.user_agent,
    }
    return Response(data=data, status_code=status.HTTP_202_ACCEPTED)
"""

models_py = """from panther.db import Model
"""

serializers_py = """from panther.serializer import ModelSerializer
"""

throttling_py = """from datetime import timedelta

from panther.throttling import Throttle

InfoThrottle = Throttle(rate=5, duration=timedelta(minutes=1))
"""

app_urls_py = """from app.apis import hello_world_api, info_api

urls = {
    '/': hello_world_api,
    'info/': info_api,
}
"""

configs_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""

from pathlib import Path

from panther.utils import load_env

BASE_DIR = Path(__name__).resolve().parent
env = load_env(BASE_DIR / '.env')

SECRET_KEY = env['SECRET_KEY']{DATABASE}{REDIS}{USER_MODEL}{AUTHENTICATION}{LOG_QUERIES}{AUTO_REFORMAT}

MIDDLEWARES = [
    {MONITORING}
]

# More Info: https://PantherPy.GitHub.io/urls/
URLs = 'core.urls.url_routing'

TIMEZONE = 'UTC'
""" % datetime.now().date().isoformat()

env = """SECRET_KEY='%s'
""" % generate_secret_key()

main_py = """from panther import Panther

app = Panther(__name__)
"""

urls_py = """from app.urls import urls as app_urls

url_routing = {
    '/': app_urls,
}
"""

git_ignore = """__pycache__/
.venv/
.idea/
logs/

.env
*.pdb
"""

requirements = """panther==%s
""" % version()

TEMPLATE = {
    'app': {
        '__init__.py': '',
        'apis.py': apis_py,
        'models.py': models_py,
        'serializers.py': serializers_py,
        'throttling.py': throttling_py,
        'urls.py': app_urls_py,
    },
    'core': {
        '__init__.py': '',
        'configs.py': configs_py,
        'urls.py': urls_py,
    },
    'main.py': main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}

single_main_py = """\"""
{PROJECT_NAME} Project (Generated by Panther on %s)
\"""
from datetime import datetime, timedelta
from pathlib import Path

from panther import Panther, status, version
from panther.app import API
from panther.request import Request
from panther.response import Response
from panther.throttling import Throttle
from panther.utils import load_env, timezone_now

BASE_DIR = Path(__name__).resolve().parent
env = load_env(BASE_DIR / '.env')

SECRET_KEY = env['SECRET_KEY']{DATABASE}{REDIS}{USER_MODEL}{AUTHENTICATION}{LOG_QUERIES}{AUTO_REFORMAT}

MIDDLEWARES = [
    {MONITORING}
]

InfoThrottle = Throttle(rate=5, duration=timedelta(minutes=1))

TIMEZONE = 'UTC'


@API()
async def hello_world_api():
    return {'detail': 'Hello World'}


@API(cache=True, throttling=InfoThrottle)
async def info_api(request: Request):
    data = {
        'panther_version': version(),
        'method': request.method,
        'query_params': request.query_params,
        'datetime_now': timezone_now().isoformat(),
        'user_agent': request.headers.user_agent,
    }
    return Response(data=data, status_code=status.HTTP_202_ACCEPTED)


url_routing = {
    '/': hello_world_api,
    'info/': info_api,
}

app = Panther(__name__, configs=__name__, urls=url_routing)
""" % datetime.now().date().isoformat()

SINGLE_FILE_TEMPLATE = {
    'main.py': single_main_py,
    '.env': env,
    '.gitignore': git_ignore,
    'requirements.txt': requirements,
}

DATABASE_PANTHERDB_PART = """

# More Info: https://PantherPy.GitHub.io/database/
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.PantherDBConnection',
        'path': BASE_DIR / 'database.{PANTHERDB_EXTENSION}',
        'encryption': {PANTHERDB_ENCRYPTION}
    }
}"""

DATABASE_MONGODB_PART = """

# More Info: https://PantherPy.GitHub.io/database/
DATABASE = {
    'engine': {
        'class': 'panther.db.connections.MongoDBConnection',
        'host': '127.0.0.1',
        'port': 27017,
        'database': '{PROJECT_NAME}'
    }
}"""

REDIS_PART = """

# More Info: https://PantherPy.GitHub.io/redis/
REDIS = {
    'class': 'panther.db.connections.RedisConnection',
    'host': '127.0.0.1',
    'port': 6379,
    'db': 0,
}"""

USER_MODEL_PART = """

# More Info: https://PantherPy.GitHub.io/configs/#user_model
USER_MODEL = 'panther.db.models.BaseUser'"""

AUTHENTICATION_PART = """

# More Info: https://PantherPy.GitHub.io/authentications/
AUTHENTICATION = 'panther.authentications.JWTAuthentication'"""

MONITORING_PART = """'panther.middlewares.monitoring.MonitoringMiddleware'"""

LOG_QUERIES_PART = """

# More Info: https://PantherPy.GitHub.io/log_queries/
LOG_QUERIES = True"""

AUTO_REFORMAT_PART = """

# More Info: https://pantherpy.github.io/configs/#auto_reformat/
AUTO_REFORMAT = True"""
